#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <libudev.h>
#include <stdbool.h>

#include "linux/nvme.h"
#include "nvme_ioctl.h"

#include "debug.h"
#include "prio.h"
#include "structs.h"
#include "sysfs.h"


#define PRIO_NAME "netappnvme"

#define PRIO_HIGH	8
#define PRIO_LOW	2
#define PRIO_ERROR	0

#define LOG_ID	0xC9
#define LOG_PAGE_SIZE 4096


__u32 get_nsid_from_path(struct path *pp)
{
	/* Buffer to hold the nsid from sysfs, 10 char (plus new line
	 * and null char) for maximum nsid value of 4294967295
	 **/
	char buf[12];
	unsigned int ctlr, nsid;
	int ret = 0;

	/* Get the controller id and local nsid */
	if (sscanf(pp->dev, "nvme%un%u", &ctlr, &nsid) < 2) {
		condlog(3, "%s prio: dev %s is not an nvme device", PRIO_NAME, pp->dev);
		return 0;
	}

	/* Get reference and get the nsid from sysfs */
	udev_device_ref(pp->udev);
	if (sysfs_attr_get_value(pp->udev, "nsid", buf, sizeof(buf)) <= 0) {
		condlog(3, "%s prio: Failed to get namespace id for %s", PRIO_NAME, pp->dev);
		goto out;
	}

	if (sscanf(buf, "%u\n", &nsid) != 1) {
		condlog(3, "%s prio: Could not parse namespace id for %s", PRIO_NAME, pp->dev);
		goto out;
	}
	ret = nsid;

	condlog(3, "%s prio: dev = %s, nsid = %u", PRIO_NAME, pp->dev, nsid);
out:
	udev_device_unref(pp->udev);	
	return ret;
}

int get_ns_info(struct path *pp, unsigned int timeout, __u32 *buf, int buflen)
{
	const char *dev = pp->dev;
	int fd = pp->fd;
	int err;
	int i;
	struct nvme_admin_cmd get_log = {
		.opcode		= nvme_admin_get_log_page,
		.nsid		= 0xffffffff,
		.addr		= (__u64)buf,
		.data_len	= buflen,
	};
	__u32 dwords = (get_log.data_len >> 2) - 1; // zero-relative, so -1
	__u32 numdu = dwords >> 16;
	__u32 numdl = dwords & 0xFFFF;

	get_log.cdw10 = LOG_ID | (numdl << 16);
	get_log.cdw11 = numdu;
	get_log.timeout_ms = get_prio_timeout(timeout, 60000); // default 60 second timeout

	condlog(3, "%s prio: dev = %s, fd = %d", PRIO_NAME, dev, fd);

	memset(buf, 0x00, buflen);

	err = ioctl(fd, NVME_IOCTL_ADMIN_CMD, &get_log);
	if (err) {
		condlog(2, "%s prio: ioctl to dev %s error %0x", PRIO_NAME, dev, err);
		return 1;
	}

#if 0
	// debug code
	for (i = 0; i < buflen / sizeof(__u32); i++) {
		/* zero is not a valid namespace id */
		if (buf[i] == 0)
			break;

		condlog(4, "%s prio: dev = %s, buf[%d] = %d", PRIO_NAME, dev, i, *(__u32 *)&buf[i]);
	}
#endif

	return 0;
}

bool check_namespace(__u32 nsid, __u32 *log_buffer, int buflen) {
	int i;

	for (i = 0; i < buflen / sizeof(__u32); i++) {
		/*
		 * Stop early if we see a zero entry (meaning we've gone past the
                 * valid entries) or if we see a namespace id greater than the 
		 * one we're looking for (the ids are in numerical order).
		 */
		if (log_buffer[i] == 0 ||
		    log_buffer[i] > nsid)	
			break;

		if (log_buffer[i] == nsid)
			return true; /* found it */
	}

	return false;
}

int getprio(struct path *pp, char *args, unsigned int timeout)
{
	const char *dev = pp->dev;
	__u32 nsid;
	int err;
	bool prioritize;
	__u32 *log_buffer;

	log_buffer = (__u32 *)malloc(LOG_PAGE_SIZE);
	if (!log_buffer) {
		condlog(2, "%s prio: malloc failed for log buffer for %s", PRIO_NAME, dev);
		goto out;
	}

	/* translate from path info to actual namespace id */
	nsid = get_nsid_from_path(pp);
	if (nsid <= 0) {
		condlog(2, "%s prio: error getting nsid for %s (%d)", PRIO_NAME, dev, nsid);
		goto out;
	}

	/* read the log page to get the namespace priority information */
	err = get_ns_info(pp, timeout, log_buffer, LOG_PAGE_SIZE);
	if (err) {
		condlog(2, "%s prio: failed to get ns info for dev %s error %0x", PRIO_NAME, dev, err);
		goto out;
	}

	/* see if nsid is in the list of priority namespaces */
	prioritize = check_namespace(nsid, log_buffer, LOG_PAGE_SIZE);
	free(log_buffer);
	if (prioritize)
		return PRIO_HIGH;
	else
		return PRIO_LOW;

out:
	if (log_buffer)
		free(log_buffer);

	return PRIO_ERROR;
}
